SETTING_FILENAME = 'filename'
SETTING_RECENT_FILES = 'recentFiles'
SETTING_WIN_SIZE = 'window/size'
SETTING_WIN_POSE = 'window/position'
SETTING_WIN_GEOMETRY = 'window/geometry'
SETTING_LINE_COLOR = 'line/color'
SETTING_FILL_COLOR = 'fill/color'
SETTING_ADVANCE_MODE = 'advanced'
SETTING_WIN_STATE = 'window/state'
SETTING_PROJECT_DIR = 'dir'
SETTING_PROJECT_FILE = 'projFile'
SETTING_PAINT_LABEL = 'paintlabel'
SETTING_LAST_OPEN_DIR = 'lastOpenDir'
SETTING_AUTO_SAVE = 'autosave'
SETTING_SINGLE_CLASS = 'singleclass'
FORMAT_PASCALVOC = 'PascalVOC'
FORMAT_YOLO = 'YOLO'
PROJECT_FILE_EXT = ".annt"
SETTING_DRAW_SQUARE = 'draw/square'
DEFAULT_ENCODING = 'utf-8'
MAX_RECENT = 7
XML_EXT = '.xml'
FILE_TREE_WIDTH_COL1 = 400
TXT_EXT = '.txt'
##Project files

PROJECT_VERIFIED_IMAGES = 'verifiedImgs'
PROJECT_NON_VERIFIED_IMAGES = 'nonVerifiedImgs'
PROJECT_CATEGORIES = 'categories'
PROJECT_ALL_IMAGES_NAMES = 'imageNames'
FILE_TREE_COLUMNS_NUMBER = 2
CONFIG_PATH = "seeds_faster/configs/seed/e2e_faster_rcnn_R_50_C4_1x_seed_strat2.yaml"

#20 colors to pick from

COLORS = ['#e6194b', '#3cb44b', '#ffe119', '#4363d8', '#f58231', 
    '#911eb4', '#46f0f0', '#f032e6', '#bcf60c', '#fabebe', '#008080',
    '#e6beff', '#9a6324', '#fffac8', '#800000', '#aaffc3', '#808000',
    '#ffd8b1', '#000075', '#808080', '#ffffff', '#000000']

##some parameters
IMAGE_DEFAULT_SCORE = 0.0
IMAGE_HUMAN_VERIFIED_SCORE = 1.0



import sys, os
from .ustr import ustr
from .image import Image
from .constants import XML_EXT, TXT_EXT
try:
    from PyQt5.QtGui import *
    from PyQt5.QtCore import *
    from PyQt5.QtWidgets import *
except ImportError:
    # needed for py3+qt4
    # Ref:
    # http://pyqt.sourceforge.net/Docs/PyQt4/incompatible_apis.html
    # http://stackoverflow.com/questions/21217399/pyqt4-qtcore-qvariant-object-instead-of-a-string
    if sys.version_info.major >= 3:
        import sip
        sip.setapi('QVariant', 2)
    from PyQt4.QtGui import *
    from PyQt4.QtCore import *
class FileLoader(object):

    def __init__(self, project):
        self.project = project

    def scanAllImages(self):
        '''scans all images in the project data file and project directories
            if images from directory are in the file, load them from the file
            otherwise, create new image object
            split images in verified, non-veriffied. Just created images are non-verified.
        '''
        # import  pdb;
        # pyqtRemoveInputHook(); pdb.set_trace()
        extensions = ['.%s' % fmt.data().decode("ascii").lower() for fmt in QImageReader.supportedImageFormats()]
        verified_images = {}
        non_verified_images = {}
        names = []
        print(self.project.all_image_names)
        for root, dirs, files in os.walk(self.project.path):
            for file in files:
                if file.lower().endswith(tuple(extensions)):
                    path = os.path.join(root, file)
                    # print(path)
                    full_path = ustr(os.path.abspath(path))
                    relative_path = os.path.relpath(full_path, self.project.path) 
                    names.append(relative_path)
                    # print("Restored all image names length is", len(self.project.all_image_names))
                    print(path)
                    if relative_path not in self.project.all_image_names:
                        print("NEW IMAGE")
                        #we have a new image which is not in the project data file 
                        new_image = Image(full_path, relative_path) 
                        non_verified_images[relative_path] = new_image 
                    else:
                        #do lookup
                        retrieved_image, is_ver = self.project.get_image_from_name(relative_path)
                        retrieved_image.path = os.path.join(self.project.path, relative_path)

                        #if project was moved, we might need to change the absolute path
                        if retrieved_image.label_file and not os.path.exists(retrieved_image.label_file.path):
                            prefix = XML_EXT if self.project.is_VOC_format else TXT_EXT
                            basename = os.path.splitext(retrieved_image.path)[0]
                            print(basename)
                            
                            #if cannot find annotation, try to change path
                            if os.path.exists(os.path.join(basename + prefix)):
                                retrieved_image.label_file.path = os.path.join(basename + prefix)
                           
                            else:
                                #otherwise, set no label file
                                retrieved_image.label_file = None

                        if retrieved_image and is_ver and retrieved_image.label_file:
                            verified_images[relative_path] = retrieved_image
                        elif retrieved_image and not is_ver:
                            non_verified_images[relative_path] = retrieved_image
                            # non_verified_images.append(retrieved_image)
                        else:

                            #there is an error/inconsistency:
                            #image was moved and label file was lost
                            new_image = Image(full_path, relative_path) 
                            non_verified_images[relative_path] = new_image 

        self.project.non_verified_images = non_verified_images
        self.project.verified_images = verified_images
        self.project.all_image_names = names
        #TODO: sort them??
        # images.sort(key=lambda x: x.lower())

    def check_xml(self, filePath):
        pass



import random
import os
import logging
import numpy as np
import csv
import pandas as pd
from .category import Category
from .lib import generate_new_color
from .file_loader import FileLoader
from .project_data import ProjectData
from .constants import *
try:
    from PyQt5.QtGui import *
    from PyQt5.QtCore import *
    from PyQt5.QtWidgets import *
except ImportError:
    if sys.version_info.major >= 3:
        import sip
        sip.setapi('QVariant', 2)
    from PyQt4.QtGui import *
    from PyQt4.QtCore import *
import collections
from collections import defaultdict
class Project(object):
    #TODO : load file from the project file
    suffix = PROJECT_FILE_EXT
    def __init__(self, path, project_name ='project', project_file = None, parent=None):
        #TODO: let user choose the name
        self.parent = parent
        self.name = project_name
        self.path = path
        #dictionaries image name --> Image object
        self._verified_images = {}
        self._non_verified_images = {}

        self.all_image_names = []
        self._num_images = 0
        self._categories = []
        self._colors = []
        self.is_VOC_format = True #otherwise should be YOLO
        self.recent_files = []
        self.file_loader = FileLoader(self)   
        self._is_saved = False    
        #currently holds just a path to the Pickle file
        self._project_file = None
        if project_file is not None and Project.is_project_file(project_file):
            self._project_file = project_file
        self._project_data = None


    @property
    def verified_images(self):
        return self._verified_images

    @verified_images.setter
    def verified_images(self, value):
        #TODO: check that required format is satisfied.
        self._verified_images = value
    
    @property
    def non_verified_images(self):
        return self._non_verified_images

    @non_verified_images.setter
    def non_verified_images(self, value):
        #TODO: check that required format is satisfied.
        self._non_verified_images = value

    @property
    def num_images(self):
        return self.num_images

    @num_images.setter
    def num_images(self, value):
        #TODO: check that required format is satisfied.
        if not isinstance(value, int):
            raise ValueError("Not an int")    
        self._num_images = value

    @property
    def categories(self):
        return self._categories

    @categories.setter
    def categories(self, value):
        #TODO: check that required format is satisfied.
        self._categories = value


    # @property
    # def is_VOC_format(self):
    #     return self._is_VOC_format

    # @is_VOC_format.setter
    # def is_VOC_format(self, value):
    #     #TODO: check that required format is satisfied.
    #     if not isinstance(value, bool):
    #         raise ValueError("Not a boolean")
    #     self._is_VOC_format = value

    @property
    def is_saved(self):
        print('returning project save state', self._is_saved)
        return self._is_saved
    
    @is_saved.setter
    def is_saved(self, value):
        if not isinstance(value, bool):
            raise ValueError("Not a boolean")
        print('chaning project save state on', value)
        self._is_saved = value


    @staticmethod
    def is_project_file(filename):
        #TODO: check that Pickle file exists
        if not os.path.exists(filename):
            return False
        fileSuffix = os.path.splitext(filename)[1].lower()
        return fileSuffix == Project.suffix
   
    ##Project  loading, saving,
    # Loading images and files
     
    def save_project_file(self):
        '''saves the project in 'name.annt' format
        returns True if saved, False otherwise'''
        # import pdb;
        # pyqtRemoveInputHook(); pdb.set_trace()
        self.store_project_data()
        if self._project_data is not None:
            save_file = os.path.join(self.path, "{}{}".format(self.name, Project.suffix))
            try:
                with open(save_file, "w") as f:
                    f.write(self._project_data.rel_path)
                    print('written a file')
                    self._is_saved = True
                    return True
            except EnvironmentError as e: # parent of IOError, OSError *and* WindowsError where available
                    print(e)
                    self._is_saved = False
                    return False
        return False
                

    #Try to load project file and data file 
    #try to load (restore) files from the project data file
    def load_project_file(self):
        print('Trying to load the project file')
        if self._project_file is not None and os.path.exists(self._project_file):
            try:
                with open(self._project_file, "r") as f:
                    last_saved = f.read()
                    last_saved = last_saved.strip()
                    # last_saved = os.path.join(self.path, last_saved)
                    data = ProjectData(self, last_saved)
                    #if valid data file
                    if data.load():
                        self._project_data = data
                        self.restore_project_data() 
                        return True                   
            except IOError:
                # if file doesn't exist, maybe because it has just been
                # deleted by a separate process
                pass
            except ValueError as e:
                print (e)
        return False

    def restore_project_data(self):
        '''restore all project data (images with label files), categories from the binary
            check for consistency with the current data in the folder
        ''' 
        if self._project_data is None:
            return
        print("restoring project data")
        self._verified_images = self._project_data.get(PROJECT_VERIFIED_IMAGES, [])
        self._non_verified_images = self._project_data.get(PROJECT_NON_VERIFIED_IMAGES, [])
        self.all_image_names = self._project_data.get(PROJECT_ALL_IMAGES_NAMES, [])
        # self._categories = self._project_data.get(PROJECT_CATEGORIES, [])
        self.check_for_consistency()

    def store_project_data(self):
        '''store all project data (images with label files), categories to the binary
            makes sure images are in the right dictionaries
        '''
        if self._project_data is None:
            self._project_data =  ProjectData(self, str(self.name) + '.pkl')
            
        self.check_for_consistency()
        self._project_data[PROJECT_VERIFIED_IMAGES]  = self._verified_images 
        self._project_data[PROJECT_NON_VERIFIED_IMAGES] = self._non_verified_images 
        self._project_data[PROJECT_ALL_IMAGES_NAMES] = self.all_image_names 
        # self._project_data[PROJECT_CATEGORIES] = self._categories     
        self._project_data.save()
    
    def check_for_consistency(self):
        '''checks that images in verified and non-verified dict have the matching label'''
        for imageName in list(self._non_verified_images):
            if self._non_verified_images[imageName].is_verified:
                print("THERE IS AN ERROR, non verified image was not marked propery")
                #there is an error, assume Image label is right
                self._verified_images[imageName] = self._non_verified_images.pop(imageName)
    

        for imageName in list(self.verified_images):
            if not self._verified_images[imageName].is_verified:
                print("THERE IS AN ERROR, verified image was not marked propery")
                #there is an error, assume Image label is right
                self._non_verified_images[imageName] = self._verified_images.pop(imageName)         
        
        
    def load_dir_images(self, window):
        def addParent(parent, column, title, data):
            item = QTreeWidgetItem(parent, [title])
            item.setData(column, Qt.UserRole, data)
            item.setChildIndicatorPolicy(QTreeWidgetItem.ShowIndicator)
            item.setExpanded (True)
            return item

        def addChild( parent, column, title, data):
            item = QTreeWidgetItem(parent, [title])
            item.setData(column, Qt.UserRole, data)
            return item

        if not self.file_loader:
            return

        window.filePath = None
        window.fileTreeWidget.clear()
        window.fileTreeWidget.setColumnCount(FILE_TREE_COLUMNS_NUMBER) #2
        parent = window.fileTreeWidget.invisibleRootItem()
        window.fileTreeWidget.setColumnWidth(0, FILE_TREE_WIDTH_COL1)
        verified_item = addParent(parent, 0, 'verified', 'V')
        non_verified_item = addParent(parent, 0, 'non verified', "NV")

        self.file_loader.scanAllImages()
        # print(self.all_image_names)
        
        for imageName in self._non_verified_images.keys():
            addChild(non_verified_item, 0, imageName, "")

        for imageName in self.verified_images.keys():
            addChild(verified_item,  0, imageName, "")

        # firstImage, _ = self.get_image_from_name(window.fileTreeWidget.topLevelItem(0).text(0))
        # window.loadImage(firstImage)
       


    def get_image_from_name(self, name):
        '''if image is in one of lists, returns the image and False if verified, otherwise, True
            if image is not in any of lists, return None and None
            assumes images were checked to be in the right dictionary
        '''
        if name in self._non_verified_images:
            return self._non_verified_images[name], False

        if name in self._verified_images:
            return self._verified_images[name], True
        
        return None, None    

    #appends a new class to the project's list if does not exist yet
    def append_category(self, new_category):
        if self._categories is not None:
            if not isinstance(new_category, Category):
                raise ValueError("Not a Class")
            if new_category in self._categories:
                #TODO notify user that class already exists
                return False
            self.assign_color(new_category)
            # print(new_category.color)
            self._categories.append(new_category)
            return True
        return False

    #assigns unique color to a class, to be persited in the data file        
    def assign_color(self, new_category):
        color = generate_new_color(self._colors)
        self._colors.append(color)
        new_category.color = color
        return new_category.color
        

    def exportCSV(self, path="stats.csv"):
        '''generates and saves CSV file will all verified images and statistics in the project directory'''
        # i = 0
        # #make sure the name is unique
        # while os.path.exists(os.path.join(self.path, f"stats_{i}.csv")):
        #     i += 1  
        # # counter_list = []
        # path = os.path.join(self.path, path)
        df_list = []
        logging.debug(f"exporting CSV to {path}")
        # writer = csv.writer(csvfile)
        for image in self._verified_images.values():
            if image.label_file:
                shapes = image.load_labels(is_VOC = self.is_VOC_format)
                if shapes:
                    cnt = collections.Counter(
                        np.array([shape[0] for shape in shapes]))
                    # cnt['image'] = image.name
                    # print(cnt)
                    df = pd.DataFrame(cnt, index = [image.name])
                    # counter_list.append(cnt)
                    df_list.append(df)
                    # for (cl, count) in cnt.most_common():
                    #     writer.writerow([image.name, [cl, count]])
                    # self.countLabel.setText([ c for c in cnt.most_common()])
        # cnt = sum(counter_list, collections.Counter())
        # for (cl, count) in cnt.most_common():
        #     writer.writerow(['total', cl, count])

        if not len(df_list) > 0:
            res = "Could not find any shape on verified images"
            logging.debug(res)
            QMessageBox.warning(self.parent, "Not exported", res)
            return res
        df = pd.concat(df_list)
        df.loc['total'] = df.sum()
        # df['percentage of germinated'] =  df['radical'] * 100.0 / (df['seed'] + df['radical'])
        df['percentage of germinated'] =  df['radical'] * 100.0 / df['seed']

        with open(path,'w') as csvfile:
            df.to_csv(csvfile)
            return f'written a file to {path}'

        return("Could not save a file")

                  



import pickle
import os
import sys


class ProjectData(object):
    def __init__(self,  project, rel_path):
        self.project  = project
        self.data = {} 
        self.rel_path = rel_path
        full_path = os.path.join(self.project.path, rel_path)
        # if not os.path.exists(full_path):
        #     raise ValueError(f'not a valid path {full_path}')
        self.path = full_path

    def __setitem__(self, key, value):
        self.data[key] = value

    def __getitem__(self, key):
        return self.data[key]

    def get(self, key, default=None):
        if key in self.data:
            return self.data[key]
        return default

    def save(self):
        if self.path:
            try:
                with open(self.path, 'wb') as f:
                    print(self.path)
                    pickle.dump(self.data, f, pickle.HIGHEST_PROTOCOL)
                    return True
            except:
                print('Writing setting failed')
        else:
            print('Save path does not exists')
        return False

    def load(self):
        try:
            print(self.path)
            if os.path.exists(self.path):
                with open(self.path, 'rb') as f:
                    print("Opening proj")
                    self.data = pickle.load(f)
                    print('Loading the data ', self.path)
                    return True
            else:
                print('Load path does not exists')
        except Exception as e:
            print('Loading setting failed ', e )
        return False

    def reset(self):
        if os.path.exists(self.path):
            os.remove(self.path)
            print('Remove setting pkl file ${0}'.format(self.path))
        self.data = {}
        self.path = None



__version_info__ = ('0', '0', '32')
__version__ = '.'.join(__version_info__)
__authors__ = "Merey Ramazanova, Silvio Giancola"
__authors_username__ = "ramazam, giancos"
__author_email__ = "merey.ramazanova@kaust.edu.sa"
__github__ = 'https://github.com/meryusha/seeds_labeler/'


import random
import os
import logging
from .category import Category
from .labelFile import LabelFile, LabelFileError
from .constants import IMAGE_DEFAULT_SCORE, IMAGE_HUMAN_VERIFIED_SCORE,XML_EXT, TXT_EXT
from .shape import Shape, DEFAULT_LINE_COLOR, DEFAULT_FILL_COLOR
from .pascal_voc_io import PascalVocReader
from .yolo_io import YoloReader

from .lib import struct, newAction, newIcon, addActions, fmtShortcut, generateColorByText
try:
    from PyQt5.QtGui import *
    from PyQt5.QtCore import *
    from PyQt5.QtWidgets import *
except ImportError:
    # needed for py3+qt4
    # Ref:
    # http://pyqt.sourceforge.net/Docs/PyQt4/incompatible_apis.html
    # http://stackoverflow.com/questions/21217399/pyqt4-qtcore-qvariant-object-instead-of-a-string
    if sys.version_info.major >= 3:
        import sip
        sip.setapi('QVariant', 2)
    from PyQt4.QtGui import *
    from PyQt4.QtCore import *

class Image(object):   
    def __init__(self,  path, name = None, label_file = None, score = IMAGE_DEFAULT_SCORE):
        self.name = name
        #name is a relative path with respect to project dir
        self.path = path
        self._is_verified = False
        # self._is_saved = True
        self.label_file = label_file
        self.score = score

    
    @property
    def is_verified(self):
        return self._is_verified
    
    # @property
    # def is_saved(self):
    #     return self._is_saved

    
    def load_labels(self, window = None, image_data = None, is_VOC = True):
        if self.label_file:
            logging.debug('TRYING TO LOAD LABELS from object')
            # self.load_shapes(window, self.label_file.shapes)
            return self.load_labels_from_file(window, self.label_file.path, is_VOC)
        else:                        
            logging.debug('TRYING TO LOAD LABELS from file')     
            if self.path is not None:
                basename = os.path.splitext(self.path)[0]
                xmlPath = os.path.join(basename + XML_EXT)
                txtPath = os.path.join( basename + TXT_EXT)
                if os.path.isfile(xmlPath):
                    return self.load_labels_from_file(window, xmlPath, True)
                elif os.path.isfile(txtPath):
                    return self.load_labels_from_file(window, txtPath, False, image_data)
        return None

    def load_labels_from_file(self, window, annotationFilePath, is_VOC = True, image_data = None, show = True):
        logging.debug("load_labels_from_file")
        if not os.path.isfile(annotationFilePath) :
            logging.debug("not os.path.isfile(annotationFilePath)")
            return None
        if is_VOC:
            parseReader = PascalVocReader(annotationFilePath)
        else: 
            if image_data:
                parseReader = YoloReader(annotationFilePath, QImage.fromData(image_data))
        if parseReader:
            shapes = parseReader.getShapes()
            if show and window:
                self.load_shapes(window, shapes)
            return shapes 
        return None
        # self.canvas.verified = parseReader.verified


    def load_shapes(self, window,  shapes):
        s = []
        print('load_shapes', len(shapes))
        # import pdb; pyqtRemoveInputHook();
        # pdb.set_trace()
        for label, points, line_color, fill_color, difficult in shapes:
            # print('line', line_color)
            shape = Shape()
            category = Category(label)
            #check if category is new and assign new color
            if window.project.append_category(category):
                shape.category = category
                color = window.project.assign_color(shape.category)
                shape.line_color = color
                shape.fill_color = color
            else:
                try:
                    shape.category = next(cat for cat in window.project.categories if cat == category)
                    shape.line_color = shape.category.color
                    shape.fill_color = shape.category.color
                except:
                    pass
            for x, y in points:
                # Ensure the labels are within the bounds of the image. If not, fix them.
                x, y, snapped = window.canvas.snapPointToCanvas(x, y)
                if snapped:
                    window.setUnsavedChanges()
                shape.addPoint(QPointF(x, y))
            shape.difficult = difficult
            shape.close()
            s.append(shape)

            if line_color:
                shape.line_color = QColor(*line_color)

            if fill_color:
                shape.fill_color = QColor(*fill_color)

            window.addLabel(shape)
        print('load_shapes, s', len(s))
        window.canvas.loadShapes(s)

    def save_labels_to_path(self, window, is_VOC):
        def format_shape(s):
            return dict(label=s.category.name,
                        line_color=s.line_color.getRgb(),
                        fill_color=s.fill_color.getRgb(),
                        points=[(p.x(), p.y()) for p in s.points],
                       # add chris
                        difficult = s.difficult)

        shapes = [format_shape(shape) for shape in window.canvas.shapes]
        # Can add differrent annotation formats here
        try:
            if is_VOC :
                if self.label_file.path[-4:].lower() != ".xml":
                    self.label_file.path += XML_EXT
                self.label_file.savePascalVocFormat(shapes, DEFAULT_LINE_COLOR.getRgb(), DEFAULT_FILL_COLOR.getRgb())               
            else:
                if self.label_file.path[-4:].lower() != ".txt":
                    self.label_file.path += TXT_EXT
                self.label_file.savesYoloFormat(shapes, [],  # self.project.categories,
                                                  DEFAULT_LINE_COLOR.getRgb(), DEFAULT_FILL_COLOR.getRgb())        
            return True
        except LabelFileError as e:
            window.errorMessage(u'Error saving label data', u'<b>%s</b>' % e)
        return False

    def save_labels(self, window, is_VOC = True):
        if not self.label_file or not self.label_file.path:
            basename = os.path.splitext(self.path)[0]
            # print(basename)
            if is_VOC:
                annotationFilePath = os.path.join(basename + XML_EXT)
            else:
                annotationFilePath = os.path.join(basename + TXT_EXT) 
            self.label_file = LabelFile(self, annotationFilePath)
            self.label_file.is_verified = self._is_verified            

        if self.save_labels_to_path(window, is_VOC):
            window.setFileSaved()
            # self._is_saved = True
            window.statusBar().showMessage('Saved to  %s' % self.label_file.path)
            window.statusBar().show()

    def verify(self):
        if self.label_file is not None:
            self._is_verified = True
        return self._is_verified
        




#!/usr/bin/env python
# -*- coding: utf-8 -*-
import codecs
import distutils.spawn
import os.path
import platform
import re
import sys
import subprocess

from time import sleep
from functools import partial
from collections import defaultdict
import logging
try:
    from PyQt5.QtGui import *
    from PyQt5.QtCore import *
    from PyQt5.QtWidgets import *
    from PyQt5.QtWidgets import QMessageBox, QFileDialog, QPushButton

except ImportError:
    # needed for py3+qt4
    # Ref:
    # http://pyqt.sourceforge.net/Docs/PyQt4/incompatible_apis.html
    # http://stackoverflow.com/questions/21217399/pyqt4-qtcore-qvariant-object-instead-of-a-string
    if sys.version_info.major >= 3:
        import sip
        sip.setapi('QVariant', 2)
    from PyQt4.QtGui import *
    from PyQt4.QtCore import *


import src.resources
# Add internal libs
from src.libs.category import Category
from src.libs.constants import *
from src.libs.lib import struct, newAction, newIcon, addActions, fmtShortcut, generateColorByText
from src.libs.settings import Settings
from src.libs.shape import Shape, DEFAULT_LINE_COLOR, DEFAULT_FILL_COLOR
from src.libs.stringBundle import StringBundle
from src.libs.canvas import Canvas
from src.libs.zoomWidget import ZoomWidget
from src.libs.labelDialog import LabelDialog
from src.libs.projectDialog import ProjectDialog
# from libs.colorDialog import ColorDialog
from src.libs.labelFile import LabelFile, LabelFileError
from src.libs.toolBar import ToolBar
from src.libs.pascal_voc_io import PascalVocReader
from src.libs.constants import XML_EXT
from src.libs.yolo_io import YoloReader
from src.libs.yolo_io import TXT_EXT
from src.libs.ustr import ustr
from src.libs.version import __version__
from src.libs.hashableQListWidgetItem import HashableQListWidgetItem
from src.libs.project import Project
from src.libs.zoom import Zoom
import collections
import numpy as np
from src.gui.ui_mainwindow import Ui_MainWindow
from PyQt5 import QtCore, QtGui, QtWidgets

__appname__ = 'SeedsLabeler'

# Utility functions and classes.

def have_qstring():
    '''p3/qt5 get rid of QString wrapper as py3 has native unicode str type'''
    return not (sys.version_info.major >= 3 or QT_VERSION_STR.startswith('5.'))

def util_qt_strlistclass():
    return QStringList if have_qstring() else list

class LabelerWindow(QMainWindow, Ui_MainWindow):


    def __init__(self, defaultDir='.', defaultPrefdefClassFile=None, defaultPrefdefColorFile=None):
        super(LabelerWindow, self).__init__()
        # QtCore.QMetaObject.connectSlotsByName(self) #TODO connect SlotsByName!

        self.setupUi(self)
        #No project selected when open
        self.project  = None
        #Index of the image currently opened
        # self.imageIndex = -1
        self.currentImage = None
        # Whether we need to save or not.
        self.unsavedFile = False
        # self.unsavedProject = False
        self.defaultDir = defaultDir
        self.loadViews()
        self.zoomNavig = Zoom(self)
        self.setupConnections()       
        self.loadSettings()

    #SETIING UP
    def setupConnections(self):

        ###MENU 
        self.help.triggered.connect(self.showTutorialDialog)
        self.showInfo.triggered.connect(self.showInfoDialog)
        self.update.triggered.connect(self.updateSoftware)
        self.resetAllAction.triggered.connect(self.resetAll)

        ###MAIN TOOLBAR
        self.openProj.triggered.connect(self.openProjectClicked)
        self.quit.triggered.connect(self.close)
        self.createProject.triggered.connect(self.createProjectClicked)
        self.saveAnnot.triggered.connect(self.saveFileClicked)
        self.saveProj.triggered.connect(self.saveProject)
        self.closeAction.triggered.connect(self.closeFile)
    
        self.verify.triggered.connect(self.verifyImg)

        ###RECTBOX
        #create Rect box pressed
        self.create.triggered.connect(self.createShape)
        #deleteShape is pressed
        self.deleteAction.triggered.connect(self.deleteSelectedShape)
        self.deleteAllLabels.clicked.connect(self.deleteAllShape)
        #copy shape is pressed
        self.copy.triggered.connect(self.copySelectedShape)
        self.edit.triggered.connect(self.editLabel)
        #hide/show rectbox
        self.hideAll.triggered.connect(partial(self.togglePolygons, False))
        self.showAll.triggered.connect(partial(self.togglePolygons, True))
        self.hideCurrentClass.triggered.connect(partial(self.toggleCurrentClassPolygons, False))
        self.showCurrentClass.triggered.connect(partial(self.toggleCurrentClassPolygons, True))


        ###BOXLABELS & LABELLIST
        self.editButton.setDefaultAction(self.edit)
        self.labelList.itemActivated.connect(self.labelSelectionChanged)
        self.labelList.itemSelectionChanged.connect(self.labelSelectionChanged)
        self.labelList.itemDoubleClicked.connect(self.editLabel)
        self.labelList.itemChanged.connect(self.labelItemChanged)
        self.moreStats.clicked.connect(self.showMoreStats)


        ###FILELIST    
        # self.fileTreeWidget.itemDoubleClicked.connect(self.fileitemDoubleClicked) 
        self.fileTreeWidget.itemDoubleClicked.connect(self.fileitemDoubleClicked)   
        self.exportCSV.clicked.connect(self.exportToCSV)         
        self.nextImage.clicked.connect(self.showNextImage)       
        self.prevImage.clicked.connect(self.showPrevImage)
        self.detectAll.clicked.connect(self.detectAllImages)
        # self.menu_file.aboutToShow.connect(self.updateFileMenu)
         # self.scrollArea = self.scroll


         ###CANVAS
        self.canvas.scrollRequest.connect(self.scrollRequest)
        self.canvas.newShape.connect(self.newShape)
        self.canvas.newShapes.connect(self.newShapes)
        self.canvas.shapeMoved.connect(self.setFileChanged)
        self.canvas.selectionChanged.connect(self.shapeSelectionChanged)
        self.canvas.drawingPolygon.connect(self.toggleDrawingSensitive)        
        self.canvas.zoomRequest.connect(self.zoomNavig.zoomRequest)

    def loadViews(self):
        self.screencastViewer = self.getAvailableScreencastViewer()
        #TODO: put our tutorial there
        self.screencast = "https://youtu.be/p0nR2YsCY_U"

        # Load string bundle for i18n
        # self.stringBundle = StringBundle.getBundle()
        # getStr = lambda strId: self.stringBundle.getString(strId)
        self.canvas = Canvas(parent=self)
        # self.colorDialog = ColorDialog(parent=self)
        self.labelDialog = LabelDialog(parent=self, listItem=[])
        self.projectDialog = ProjectDialog(parent = self)
        self.itemsToShapes = {}
        self.shapesToItems = {}
        self.prevLabelText = ''
        self.lineColor = None
        self.fillColor = None         

         # what is it? 
        self._noSelectionSlot = False
        self.scroll.setWidget(self.canvas)
        # scroll.setWidgetResizable(True)
        self.scrollBars = {
            Qt.Vertical: self.scroll.verticalScrollBar(),
            Qt.Horizontal: self.scroll.horizontalScrollBar()
        }
       

        self.labelMenu = QMenu()
        self.context = (self.edit, self.deleteAction, self.hideCurrentClass, self.showCurrentClass, self.hideAll, self.showAll)
        addActions(self.labelMenu, self.context)
        addActions(self.canvas.menus[0], self.context)

        # right click drag
        action = partial(newAction, self)
        addActions(self.canvas.menus[1], (
                    action('&Copy here', self.copyShape),
                    action('&Move here', self.moveShape)))


        self.labelList.setContextMenuPolicy(Qt.CustomContextMenu)
        self.labelList.customContextMenuRequested.connect(
            self.popLabelListMenu)



        self.onLoadActive = (self.closeAction, self.create, self.detect, self.detectAll,
                          self.createMode, self.editMode)

        # Active actions when Shape are present
        self.onShapesPresent = (self.saveAs, self.hideAll, self.showAll)
        
        
         
      
        self.lastLabel = None
       
        self.displayLabelOption.triggered.connect(self.togglePaintLabelsOption)

        # STATUS BAR
        self.statusBar().showMessage('%s started.' % __appname__)
        self.statusBar().show()

        # Populate the File menu dynamically.
        # self.updateFileMenu()

        # Display cursor coordinates at the right of status bar
        self.labelCoordinates = QLabel('')
        self.statusBar().addPermanentWidget(self.labelCoordinates)

    def loadSettings(self):
        # Load setting in the main thread
        self.settings = Settings()
        self.settings.load()
        # Application state.

        projFile = ustr(self.settings.get(SETTING_PROJECT_FILE, None))
        if self.project is None and self.openProject(projFile):
            if self.project is not None:
                self.statusBar().showMessage('%s started. Annotation will be saved to %s' %
                                            (__appname__, self.project.path))
                self.statusBar().show()
                # import pdb;
                # pyqtRemoveInputHook(); pdb.set_trace()

                self.autoSaving.setChecked(self.settings.get(SETTING_AUTO_SAVE, False))
                self.singleClassMode.setChecked(self.settings.get(SETTING_SINGLE_CLASS, False))
                self.displayLabelOption.setChecked(self.settings.get(SETTING_PAINT_LABEL, False))

                ## Fix the compatible issue for qt4 and qt5. Convert the QStringList to python list
                if self.settings.get(SETTING_RECENT_FILES):
                    if have_qstring():
                        recentFileQStringList = self.settings.get(SETTING_RECENT_FILES)
                        self.project.recent_files = [ustr(i) for i in recentFileQStringList]
                    else:
                        self.project.recent_files = recentFileQStringList = self.settings.get(SETTING_RECENT_FILES)

                size = self.settings.get(SETTING_WIN_SIZE, QSize(600, 500))
                position = QPoint(0, 0)
                saved_position = self.settings.get(SETTING_WIN_POSE, position)
                # Fix the multiple monitors issue
                for i in range(QApplication.desktop().screenCount()):
                    if QApplication.desktop().availableGeometry(i).contains(saved_position):
                        position = saved_position
                        break
                self.resize(size)
                self.move(position)

                # self.lastOpenDir = ustr(self.settings.get(SETTING_LAST_OPEN_DIR, None))

                self.restoreState(self.settings.get(SETTING_WIN_STATE, QByteArray()))
                Shape.line_color = self.lineColor = QColor(self.settings.get(SETTING_LINE_COLOR, DEFAULT_LINE_COLOR))
                Shape.fill_color = self.fillColor = QColor(self.settings.get(SETTING_FILL_COLOR, DEFAULT_FILL_COLOR))
                self.canvas.setDrawingColor(self.lineColor)

    
    #Project creation, opening, saving, closing
    def createProjectClicked(self, _value=False, dirpath=None):
        if not self.mayContinueWithProject():
            return

        defaultOpenDirPath = dirpath if dirpath else '.'
        targetDirPath = ustr(QFileDialog.getExistingDirectory(self, '%s - Open Directory' % __appname__, defaultOpenDirPath,
                                                     QFileDialog.ShowDirsOnly | QFileDialog.DontResolveSymlinks))
        if targetDirPath:
            self.project = Project(targetDirPath, parent=self)
            self.project.load_dir_images(self)
            self.saveProj.setEnabled(True)
      
    def openProjectClicked(self, _value=False):
        '''Triggered by button pressing
            Opens a dialog and passing selected folder to openProject() method
        '''
        if not self.mayContinueWithProject():
            return
        path = os.path.dirname(ustr(self.defaultDir)) if self.defaultDir else '.'
        # formats = ['*.%s' % fmt.data().decode("ascii").lower() for fmt in QImageReader.supportedImageFormats()]
        filters = "Project files (%s)" % ' '.join(['*%s' % Project.suffix])
        filename = QFileDialog.getOpenFileName(self, '%s - Choose Project file' % __appname__, path, filters)
        if filename:
            if isinstance(filename, (tuple, list)):
                filename = filename[0]
            self.openProject(filename)
            # self.loadImage(filename)

    def openProject(self, filePath=None):
        '''Trying to load project from the project .annt file
            shows error if file is not valid, then cleans the workspace and loads the images and project settings
        '''
        print("Selecting project file")

        if filePath is None:
            return
        self.resetState()
        self.canvas.setEnabled(False)
      
        # Make sure that filePath is a regular python string, rather than QString
        filePath = ustr(filePath)
        unicodeFilePath = ustr(filePath)

        if unicodeFilePath and os.path.exists(unicodeFilePath):
            if Project.is_project_file(unicodeFilePath):
                #TODO: change path to the current dir
                project = Project(os.path.dirname(filePath), project_file = unicodeFilePath)
                if project.load_project_file():
                    self.project = project 
                    self.status("Loaded %s" % os.path.basename(unicodeFilePath))
                    self.project.load_dir_images(self)
                    self.project.is_saved = True
                    self.exportCSV.setEnabled(True)

                    return True
                else:                  
                    self.errorMessage(u'Error opening project',
                                    (u"<p>Could open <i>%s</i>. But could not restore the files")
                                    % (unicodeFilePath))
                    self.status("Error reading %s" % unicodeFilePath)
                    return False
            else:
                self.errorMessage(u'Error opening project',
                                      (u"<p>Make sure <i>%s</i> is a valid project file.")
                                      % ( unicodeFilePath))   
        return False
    
    def saveProject(self):
        self.saveFile()
        if self.project.save_project_file():
            self.setProjectSaved()
        else: 
            self.errorMessage(u'Error saving project', 'Could not save the project file')
            self.status("Error saving the project")
                         
    #File creation, opening, saving, closing
  
    def saveFileClicked(self, _value=False):
        if self.project.path is not None and len(ustr(self.project.path)):
            self.saveFile()
            self.saveProj.setEnabled(True)
        #TODO : throw and error if not saved

    def saveFile(self):
        if self.currentImage:
            self.currentImage.save_labels(self, self.project.is_VOC_format)

    def closeFile(self, _value=False):
        if not self.mayContinueWithFile():
            return
        self.resetState()
        self.setFileSaved()
        self.toggleActions(False)
        self.canvas.setEnabled(False)
        self.saveAs.setEnabled(False)
 
    ##STATE
    def resetState(self):
        self.itemsToShapes.clear()
        self.shapesToItems.clear()
        self.labelList.clear()
        self.currentImage = None
        # self.imageData = None
        # self.labelFile = None
        self.canvas.resetState()
        self.labelCoordinates.clear()

    def setFileChanged(self):
        '''
            Being called by signal when canvas shapes are moved
            Meaning both file and project have to be saved
        '''
        print('file is being changed')
        if self.project is not None:
            self.project.is_saved = False
        self.unsavedFile = True
        self.saveAnnot.setEnabled(True)
        self.saveProj.setEnabled(True)
    
   
    def setFileSaved(self):
        '''
            Called after a file is saved or closed
            Still might need to save a project
        '''
        self.unsavedFile = False
        self.saveAnnot.setEnabled(False)
        if self.currentImage:
            self.create.setEnabled(True)
        # self.detect.setEnabled(True)

    def setProjectSaved(self):
        '''Called when project was just saved
        '''
        self.setFileSaved()
        self.saveProj.setEnabled(False)


    def toggleActions(self, value=True):
        """Enable/Disable widgets which depend on an opened image."""
        for z in self.zoomNavig.zoomActions:
            z.setEnabled(value)
        for action in self.onLoadActive:
            action.setEnabled(value)

    def queueEvent(self, function):
        QTimer.singleShot(0, function)

    def status(self, message, delay=5000):
        self.statusBar().showMessage(message, delay)

   
    def currentItem(self):
        items = self.labelList.selectedItems()
        if items:
            return items[0]
        return None

    def addRecentFile(self, filePath):
        if self.project is not None:
            if filePath in self.project.recent_files:
                self.project.recent_files.remove(filePath)
            elif len(self.project.recent_files) >= MAX_RECENT:
                self.project.recent_files.pop()
            self.project.recent_files.insert(0, filePath)

    def getAvailableScreencastViewer(self):
        osName = platform.system()

        if osName == 'Windows':
            return ['C:\\Program Files\\Internet Explorer\\iexplore.exe']
        elif osName == 'Linux':
            return ['xdg-open']
        elif osName == 'Darwin':
            return ['open', '-a', 'Safari']

    ## Callbacks ##
    def showTutorialDialog(self):
        subprocess.Popen(self.screencastViewer + [self.screencast])

    def showInfoDialog(self):
        msgBox = QMessageBox()
        msgBox.setWindowTitle('Information')
        msgBox.setText(f'Name:{__appname__} \nApp Version:{__version__} \nPython:{sys.version_info}')
        msgBox.setIcon(QMessageBox.Information);
        msgBox.addButton(QPushButton('Update'), QMessageBox.YesRole)
        msgBox.addButton(QMessageBox.Ok)
        ret = msgBox.exec_()

        logging.debug(f"QMessageBox returned value {ret}, if 0, then update")
        
        if ret == 0:
            self.updateSoftware()




    def updateSoftware(self):

        CompletedProcess = subprocess.run(["pip", "list", "-o"],
                                          check=True,
                                          stdout=subprocess.PIPE, 
                                          universal_newlines=True)
        logging.debug(CompletedProcess.stdout)

        if "SeedsLabeler" in CompletedProcess.stdout:
            wantUpdate = QMessageBox.question(self, 
                "Update Available",
                "A new version of SeedsLabeler is available.\nDo you want to update it?", 
                QMessageBox.Yes | QMessageBox.No, QMessageBox.No)

            if wantUpdate == QMessageBox.Yes:
                logging.info("Updating through pip")
                subprocess.run(["pip", "install", "--upgrade", "SeedsLabeler"])

                wantRestart = QMessageBox.question(self,
                                                   'Update Done',
                                                   "Update done. Do you want to restart SeedsLabeler?\nPlease make sure you saved your project.",
                                                QMessageBox.Yes | QMessageBox.No, QMessageBox.No)
                if wantRestart == QMessageBox.Yes:
                    logging.info("Restarting SeedsLabeler")
                    import atexit
                    atexit.register(os.execl, "SeedsLabeler.py", "SeedsLabeler.py")
                    self.close()
            else:
                logging.info('New version avalaible but will not update it yet.')
        else:
            QMessageBox.information(self, 
                    'No update available', 
                    "Your software is already up-to-date.")






    def createShape(self):
        # assert self.beginner()
        self.canvas.setEditing(False)
        self.create.setEnabled(False)
        self.detect.setEnabled(False)
        self.detectAll.setEnabled(False)
        self.updateCount()


    def toggleDrawingSensitive(self, drawing=True):
        """In the middle of drawing, toggling between modes should be disabled."""
        self.editMode.setEnabled(not drawing)
        if not drawing:
            # Cancel creation.
            print('Cancel creation.')
            self.canvas.setEditing(True)
            self.canvas.restoreCursor()
            self.create.setEnabled(True)
            self.detect.setEnabled(True)
            self.detectAll.setEnabled(True)

    def toggleDrawMode(self, edit=True):
        self.canvas.setEditing(edit)
        self.createMode.setEnabled(edit)
        self.editMode.setEnabled(not edit)

    def popLabelListMenu(self, point):
        logging.debug(f"pop label at point {point}")
        # self.menu_Edit.exec_(self.labelList.mapToGlobal(point))
        self.labelMenu.exec_(self.labelList.mapToGlobal(point))

    def editLabel(self):
        if not self.canvas.editing():
            return
            
        item = self.currentItem()
        text = self.labelDialog.popUp(item.text())
        if text is not None:
            item.setText(text)
            item.setBackground(generateColorByText(text))
            self.setFileChanged()


    def togglePolygons(self, value):
        if not self.canvas.editing():
            return

        self.labelList.itemChanged.disconnect(self.labelItemChanged)

        shapes = []
        for item, shape in self.itemsToShapes.items():
            item.setCheckState(Qt.Checked if value else Qt.Unchecked)
            shapes.append(shape)

        self.canvas.setShapesVisible(shapes, value)
        self.canvas.update()
        self.canvas.repaint()
        self.labelList.itemChanged.connect(self.labelItemChanged)

    def toggleCurrentClassPolygons(self, value):
        if not self.canvas.editing():
            return

        self.labelList.itemChanged.disconnect(self.labelItemChanged)

        shapes = []
        for item, shape in self.itemsToShapes.items():
            if self.currentItem() is not None:
                if item.text() == self.currentItem().text():
                    item.setCheckState(Qt.Checked if value else Qt.Unchecked)
                    shapes.append(shape)

        self.canvas.setShapesVisible(shapes, value)
        self.canvas.update()
        self.canvas.repaint()
        self.labelList.itemChanged.connect(self.labelItemChanged)


    # Tzutalin 20160906 : Add file list and dock to move faster
    def fileitemDoubleClicked(self, item=None, col = 0):
        #Merey : In simple terms, index() method finds the given element in a list and returns its position.     
        if not self.mayContinueWithFile():
            return
        imageName = item.text(col)
        if imageName in self.project.all_image_names:
            self.index = self.project.all_image_names.index(imageName)
        # if self.index < len(self.project.all_image_names):
            image, isVerified = self.project.get_image_from_name(imageName)
            self.loadImage(image, index = self.index)

            self.nextImage.setEnabled(True)
            self.prevImage.setEnabled(True)


    def detectAllImages(self):
        logging.debug("detectAllImages")
        self.detectAllProgressBar.setMaximum(len(self.project.all_image_names))
        for new_index in range(len(self.project.all_image_names)):
        # current_index = self.project.all_image_names.index(os.path.basename(self.currentImage.path))
        # new_index = min(current_index + 1, len(self.project.all_image_names)-1)

            self.index = self.project.all_image_names.index(self.project.all_image_names[new_index])
            image, isVerified = self.project.get_image_from_name(self.project.all_image_names[new_index])
            self.loadImage(image, index = self.index)

            if not isVerified:
                self.deleteAllShape()
                # logging.debug("BEFORE EMITTING")
                self.detect.triggered.emit()
                # logging.debug("AFTER EMITTING")
                self.saveFileClicked()
                self.detectAllProgressBar.setValue(new_index+1)
                # api.detectObjects(win.currentImage.path)




    def showNextImage(self):
        logging.debug("showNextImage")
        current_index = self.project.all_image_names.index(os.path.basename(self.currentImage.path))
        new_index = min(current_index + 1, len(self.project.all_image_names)-1)

        self.index = self.project.all_image_names.index(self.project.all_image_names[new_index])
        image, isVerified = self.project.get_image_from_name(self.project.all_image_names[new_index])
        self.loadImage(image, index = self.index)
        # fileTreeWidget.setCurrentItem(fileTreeWidget.it)
        

    def showPrevImage(self):
        logging.debug("showPrevImage")
        current_index = self.project.all_image_names.index(os.path.basename(self.currentImage.path))
        new_index = max(current_index - 1, 0)

        self.index = self.project.all_image_names.index(self.project.all_image_names[new_index])
        image, isVerified = self.project.get_image_from_name(self.project.all_image_names[new_index])
        self.loadImage(image, index = self.index)
        # fileTreeWidget.setCurrentItem(fileTreeWidget.it)
        

    # React to canvas signals.
    def shapeSelectionChanged(self, selected=False):
        if self._noSelectionSlot:
            self._noSelectionSlot = False
        else:
            shape = self.canvas.selectedShape
            if shape:
                # import pdb;
                # pyqtRemoveInputHook(); pdb.set_trace()                 
                self.shapesToItems[shape].setSelected(True)
            else:
                self.labelList.clearSelection()
        self.deleteAction.setEnabled(selected)
        self.copy.setEnabled(selected)
        self.edit.setEnabled(selected)
        self.showCurrentClass.setEnabled(selected)
        self.hideCurrentClass.setEnabled(selected)
        self.shapeLineColor.setEnabled(selected)
        self.shapeFillColor.setEnabled(selected)
        self.updateCount()

    def addLabel(self, shape):
        shape.paintLabel = self.displayLabelOption.isChecked()
        item = HashableQListWidgetItem(shape.category.name)
        item.setFlags(item.flags() | Qt.ItemIsUserCheckable)
        item.setCheckState(Qt.Checked)
        print(shape.category.color)
        item.setBackground(shape.category.color)
        self.itemsToShapes[item] = shape
        self.shapesToItems[shape] = item
        self.labelList.addItem(item)
        for action in self.onShapesPresent:
            action.setEnabled(True)
        self.updateCount()

    def updateCount(self):
        # print(np.array([shape.category.name for shape in self.canvas.shapes]))
        cnt = collections.Counter(
            np.array([shape.category.name for shape in self.canvas.shapes]))
        # self.countLabel.setText([ c for c in cnt.most_common()])
        # print(cnt)
        countString = [f"{cl} : {count} \n" for (cl, count) in cnt.most_common()]
        # self.countLabel.setText("".join(str(c) for c in cnt.most_common()))
        # print(countString)
        self.countLabel.setText("".join(c for c in countString ))

    def remLabel(self, shape):
        if shape is None:
            # print('rm empty label')
            return
        item = self.shapesToItems[shape]
        self.labelList.takeItem(self.labelList.row(item))
        del self.shapesToItems[shape]
        del self.itemsToShapes[item]
        self.updateCount()

    def copySelectedShape(self):
        try:
            self.addLabel(self.canvas.copySelectedShape())
            # fix copy and delete
            self.shapeSelectionChanged(True)
        except:
            print("Error while duplicating shape")

    def labelSelectionChanged(self):
        item = self.currentItem()
        if item and self.canvas.editing():
            self._noSelectionSlot = True
            self.canvas.selectShape(self.itemsToShapes[item])
            shape = self.itemsToShapes[item]


    def labelItemChanged(self, item):
        shape = self.itemsToShapes[item]
        label = item.text()
        if label != shape.category.name:
            shape.category.name = item.text()
            shape.line_color = shape.category.color
            self.setFileChanged()
        else:  # User probably changed item visibility
            self.canvas.setShapeVisible(shape, item.checkState() == Qt.Checked)


    def showMoreStats(self):

        cnt = collections.Counter(
            np.array([shape.category.name for shape in self.canvas.shapes]))

        msg = QMessageBox()
        msg.setIcon(QMessageBox.Information)

        if cnt['seed'] > 0:
            msg.setText(f"ratio germinated/non-germinated: \n {cnt['radical']*100/cnt['seed']:02.2f} %")
        else:
            msg.setText(f"ratio germinated/non-germinated: \n NaN %")
        msg.setInformativeText("Show details for more")
        msg.setWindowTitle("More Statistics")

        detailledMessage = "The details are as follows: \n"
        detailledMessage += f" - seeds: {cnt['seed']}\n"
        detailledMessage += f" - radical: {cnt['radical']}\n"
        detailledMessage += f" - germinated seeds: {cnt['radical']}\n"
        detailledMessage += f" - non-germinated seeds: {cnt['seed']-cnt['radical']}\n"
        if cnt['seed'] > 0:
            detailledMessage += f" - ratio germinated: {cnt['radical']*100/cnt['seed']:02.2f} %\n"
        else:
            detailledMessage += f" - ratio germinated: NaN %\n"        
        msg.setDetailedText(detailledMessage)

        retval = msg.exec_()


    #params: tuple of (boxes, labels) 
    def addShapes(self, box_label_tup):
        boxes, labels, scores = box_label_tup
        # add all shapes
        for box, label, score in zip(boxes, labels, scores):
            # self.addShape(box, label)
            # TODO add all shape at once
            # print(label)
            minX, minY, maxX, maxY  = box
            self.canvas.current = Shape()
            category = Category(label)
            #check if category is new and assign new color
            # import pdb;
            # pyqtRemoveInputHook(); pdb.set_trace()
            if self.project.append_category(category):
                self.canvas.current.category = category
                color = self.project.assign_color(self.canvas.current.category)
                self.canvas.current.line_color = color
                self.canvas.current.fill_color = color
            else:
                try:
                    self.canvas.current.category = next(cat for cat in self.project.categories if cat == category)
                    self.canvas.current.line_color = self.canvas.current.category.color
                    self.canvas.current.fill_color = self.canvas.current.category.color
                except:
                    pass
            print(self.canvas.current.category)                    
            self.canvas.current.addPoint(QPointF(minX, minY))
            self.canvas.current.addPoint(QPointF(maxX, minY))
            self.canvas.current.addPoint(QPointF(maxX, maxY))
            self.canvas.current.addPoint(QPointF(minX, maxY))
            # self.canvas.current.label = str(label)
            self.canvas.current.close()
            self.canvas.shapes.append(self.canvas.current)

        # repaint only once
        self.canvas.current = None
        self.canvas.newShapes.emit(len(boxes))
        self.canvas.update()
        self.canvas.repaint()

    def addShape(self, box, label):
        # import pdb;
        # pyqtRemoveInputHook(); pdb.set_trace()
        minX = box[0]
        minY = box[1]
        maxX = box[2]
        maxY = box[3]
        self.canvas.current = Shape()
        category = Category(label)
            #check if category is new and assign new color
        if self.project.append_category(category):
            self.canvas.current.category = category
            color = self.project.assign_color(self.canvas.current.category)
            self.canvas.current.line_color = color
            self.canvas.current.fill_color = color
        else:
            try:
                self.canvas.current.category = next(cat for cat in self.project.categories if cat == category)
                self.canvas.current.line_color = self.canvas.current.category.color
                self.canvas.current.fill_color = self.canvas.current.category.color
            except:
                pass        
        self.canvas.current.addPoint(QPointF(minX, minY))
        self.canvas.current.addPoint(QPointF(maxX, minY))
        self.canvas.current.addPoint(QPointF(maxX, maxY))
        self.canvas.current.addPoint(QPointF(minX, maxY))
        # self.canvas.current.category = label
        self.canvas.current.close()

        self.canvas.shapes.append(self.canvas.current)
        self.canvas.current = None
        self.canvas.newShape.emit()
        self.canvas.update()
        self.canvas.repaint()
        return

    # Callback functions:
    def newShapes(self, n):
        """Pop-up and give focus to the label editor.
        position MUST be in global coordinates.
        """
        for i in range(n):
            shape = self.canvas.shapes[-1 - i]
            if shape.category.name is not None:
                text = shape.category.name
            elif not self.useDefaultLabelCheckbox.isChecked() or not self.defaultLabelTextLine.text():
                if len(self.project.categories) > 0:
                    self.labelDialog = LabelDialog(
                        parent=self, listItem=self.project.categories)

                # Sync single class mode from PR#106
                if self.singleClassMode.isChecked() and self.lastLabel:
                    text = self.lastLabel
                else:
                    text = self.labelDialog.popUp(text=self.prevLabelText)
                    self.lastLabel = text
            else:
                text = self.defaultLabelTextLine.text()

            # Add Chris
            # self.diffcButton.setChecked(False)
            if text is not None:
                self.prevLabelText = text

                assert text
                category = Category(text)
                #check if category is new and assign new color
                if self.project.append_category(category):
                    shape.category = category
                    shape.line_color = shape.fill_color =  self.project.assign_color(shape.category)
                else:
                    try:
                        shape.category = next(cat for cat in self.project.categories if cat is category)
                    except:
                        pass

                self.addLabel(shape)

            else:
                # self.canvas.undoLastLine()
                self.canvas.resetAllLines()

        # if self.beginner():  # Switch to edit mode.
        self.canvas.setEditing(True)
        self.create.setEnabled(True)
        self.detect.setEnabled(True)
        self.detectAll.setEnabled(True)
        # else:
        #     self.actions.editMode.setEnabled(True)
        self.setFileChanged()

    # Callback functions:
    def newShape(self):
        # import pdb
        # InputHook(); pdb.set_trace()
        """Pop-up and give focus to the label editor.
        position MUST be in global coordinates.
        """
        # print(self.canvas.shapes[-1].label)
        if self.canvas.shapes[-1].category is not None:
            text = self.canvas.shapes[-1].category
        elif not self.useDefaultLabelCheckbox.isChecked() or not self.defaultLabelTextLine.text():
            if len(self.project.categories) > 0:
                self.labelDialog = LabelDialog(
                    parent=self, listItem=self.project.categories)

            # Sync single class mode from PR#106
            if self.singleClassMode.isChecked() and self.lastLabel:
                text = self.lastLabel
            else:
                text = self.labelDialog.popUp(text=self.prevLabelText)
                self.lastLabel = text
        else:
            text = self.defaultLabelTextLine.text()

        # Add Chris
        # self.diffcButton.setChecked(False)

        if text is not None:
            self.prevLabelText = text
            category = Category(text)
            shape = self.canvas.setLastLabel(text)
            #check if category is new and assign new color
            if self.project.append_category(category):
                shape.category = category
                shape.line_color = shape.fill_color =  self.project.assign_color(shape.category)
            else:
                try:
                    shape.category = next(cat for cat in self.project.categories if cat == category)
                except:
                    pass

            self.addLabel(shape)
            self.canvas.setEditing(True)
            self.create.setEnabled(True)
            self.detect.setEnabled(True)
            self.detectAll.setEnabled(True)
            self.setFileChanged()

        else:
            # self.canvas.undoLastLine()
            self.canvas.resetAllLines()

    def scrollRequest(self, delta, orientation):
        units = - delta / (8 * 15)
        bar = self.scrollBars[orientation]
        bar.setValue(bar.value() + bar.singleStep() * units)





    def loadImage(self, image=None, index = 0):
        '''loading image data from the path and annotations (if available) from the image object,
            or loading a new one
        '''
        # import pdb;
        # pyqtRemoveInputHook(); pdb.set_trace() 
        """Load the specified file, or the last opened file if None."""
        self.resetState()
        self.canvas.setEnabled(False)
        if image is None:
            return False
        # import pdb;
        # pyqtRemoveInputHook(); pdb.set_trace()
        self.currentImage = image
        # Make sure that filePath is a regular python string, rather than QString
        unicodeFilePath = ustr(self.currentImage.path)
        if unicodeFilePath and os.path.exists(unicodeFilePath):       
            # Load image:
            # read data first and store for saving into label file.
            imageData = read(unicodeFilePath, None)
            imageFromData = QImage.fromData(imageData)
            if imageFromData.isNull():
                self.errorMessage(u'Error opening file',
                                  u"<p>Make sure <i>%s</i> is a valid image file." % unicodeFilePath)
                self.status("Error reading %s" % unicodeFilePath)
                return False
            self.status("Loaded %s" % os.path.basename(unicodeFilePath))
            self.canvas.loadPixmap(QPixmap.fromImage(imageFromData))

            self.setFileSaved()
            self.canvas.setEnabled(True)
            self.canvas.verified = self.currentImage.is_verified
            self.zoomNavig.adjustScale(initial=True)
            self.zoomNavig.paintCanvas()
            self.addRecentFile(unicodeFilePath)
            self.toggleActions(True)

            #Merey: Load Xml if available
            self.currentImage.load_labels(self, imageData, self.project.is_VOC_format)
            self.setWindowTitle(__appname__ + ' ' + unicodeFilePath)
            self.verify.setEnabled(not self.currentImage.is_verified)
            # Default : select last item if there is at least one item
            if self.labelList.count():
                self.labelList.setCurrentItem(self.labelList.item(self.labelList.count()-1))
                self.labelList.item(self.labelList.count()-1).setSelected(True)

            self.canvas.setFocus(True)
            return True
        return False


    def closeEvent(self, event):
        if not self.mayContinueWithProject():
            event.ignore()
        # settings = self.settings
        # If it loads images from dir, don't load it at the begining
        if self.project is not None:

            self.settings[SETTING_WIN_SIZE] = self.size()
            self.settings[SETTING_WIN_POSE] = self.pos()
            self.settings[SETTING_WIN_STATE] = self.saveState()
            self.settings[SETTING_LINE_COLOR] = self.lineColor
            self.settings[SETTING_FILL_COLOR] = self.fillColor
            self.settings[SETTING_RECENT_FILES] = self.project.recent_files
            # self.settings[SETTING_ADVANCE_MODE] = not self._beginner
            if self.project.path and os.path.exists(self.project.path):
                self.settings[SETTING_PROJECT_DIR] = ustr(self.project.path)
            else:
                self.settings[SETTING_PROJECT_DIR] = ''

            self.settings[SETTING_AUTO_SAVE] = self.autoSaving.isChecked()
            self.settings[SETTING_SINGLE_CLASS] = self.singleClassMode.isChecked()
            self.settings[SETTING_PAINT_LABEL] = self.displayLabelOption.isChecked()
            # self.settings[SETTING_DRAW_SQUARE] = self.drawSquaresOption.isChecked()
            self.settings.save()

    def loadRecent(self, filename):
        if self.mayContinueWithFile():
            self.loadImage(filename)    

    def verifyImg(self, _value=False):
        '''will save and mark image as verified'''
        if self.currentImage is not None:
            self.saveFileClicked()
            if self.currentImage.verify():
                self.verify.setEnabled(not self.currentImage.is_verified)
                self.canvas.verified = self.currentImage.is_verified
                if self.project.non_verified_images[self.currentImage.name] is not None:
                    self.project.non_verified_images.pop(self.currentImage.name)
                else:
                    print("Error: image does not exists in the list")
                self.project.verified_images[self.currentImage.name] = self.currentImage 
                self.project.load_dir_images(self)
                self.zoomNavig.paintCanvas()
            else:
                self.errorMessage(u'Error verifying',
                                  u"<p>Make sure you labeled an image." )
    
    def exportToCSV(self):
        logging.debug("exportToCSV")
        # logging.debug(self.project)
        if self.project is None or not len(self.project._verified_images.values()) > 0: # inconsistency - button should not be enabled
            # self.exportCSV.setEnabled(False)
            self.errorMessage(u'Error exporting',
            u"<p>Sorry, could not export to CSV.\nDid you open a project already?\nHave you verified any images?" )
        else:
            path = QtWidgets.QFileDialog.getSaveFileName(self, 'Export CSV', self.project.path, "CSV Files (*.csv)")[0]

            mes = self.project.exportCSV(path=path)
            self.status(mes)

    def resetAll(self):
        self.settings.reset()
        self.close()
        proc = QProcess()
        proc.startDetached(os.path.abspath(__file__))

    def mayContinueWithFile(self):
        return not (self.unsavedFile and not self.discardChangesDialog('if file'))

    def mayContinueWithProject(self):
        # print(self.project.is_saved)
        return not self.project or self.project.is_saved or self.discardChangesDialog("in project")

    def discardChangesDialog(self, mes):
        yes, no = QMessageBox.Yes, QMessageBox.No
        msg = u'You have unsaved changes %s, proceed anyway?' % mes 
        return yes == QMessageBox.warning(self, u'Attention', msg, yes | no)

    def errorMessage(self, title, message):
        return QMessageBox.critical(self, title,
                                    '<p><b>%s</b></p>%s' % (title, message))

    def deleteSelectedShape(self):
        # import pdb;
        # pyqtRemoveInputHook(); pdb.set_trace()        
        self.remLabel(self.canvas.deleteSelected())
        self.setFileChanged()
        if not self.itemsToShapes:
            for action in self.onShapesPresent:
                action.setEnabled(False)
    
    def deleteAllShape(self):
        # import pdb;
        # pyqtRemoveInputHook(); pdb.set_trace()  
        while len(self.canvas.shapes) > 0:
            self.canvas.selectedShape = self.canvas.shapes[0]
            self.deleteSelectedShape()
        self.canvas.update()
        self.canvas.repaint()

        

       #EVENTS HANDLING
   
    def keyReleaseEvent(self, event):
        if event.key() == Qt.Key_Control:
            self.canvas.setDrawingShapeToSquare(False)

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Control:
            # Draw rectangle if Ctrl is pressed
            self.canvas.setDrawingShapeToSquare(True)

    def copyShape(self):
        self.canvas.endMove(copy=True)
        self.addLabel(self.canvas.selectedShape)
        self.setFileChanged()

    def moveShape(self):
        self.canvas.endMove(copy=False)
        self.setFileChanged()

    def togglePaintLabelsOption(self):
        for shape in self.canvas.shapes:
            shape.paintLabel = self.displayLabelOption.isChecked()

# def inverted(color):
#     return QColor(*[255 - v for v in color.getRgb()])


def read(filename, default=None):
    try:
        with open(filename, 'rb') as f:
            return f.read()
    except:
        return default



import socket
import threading
import json
import requests
import pickle
import time
import threading 
import select

class SingleThreadedClient(object):
    def __init__(self, host, port):
        self.host = host
        self.port = port
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.settimeout(30)
        self.buffer_size = 16384
        # self.sock.setblocking(0)   
        # self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
         
    def connectToServer(self):                 
        # connect to the server on local computer
        try: 
            print(self.host)
            self.sock.connect((self.host, self.port))

            return True
        except Exception as e: 
            print(e)
            return False

    def processImageLabels(self, data):
        # print(image_path)
        return pickle.loads(data)
 
    def sendDetectionImage(self, image_path):
        try:
            print('Trying to send Image to server')
            if image_path is None:
                return
            im = open(image_path, 'rb')
            image_bytes = im.read()
            size = len(image_bytes)
            print(size)
            request = f"DETECT SIZE={size}"
            self.sock.send(request.encode('utf-8'))
            # receive data from the server (buffer size)
            response = self.sock.recv(self.buffer_size)
            if not response:
                #connection lost
                self.sock.close() 
                print('Connection lost? ') 
                return None
            if str(response.decode()).startswith("OK"):
                print('sending image bytes')
                self.sock.sendall(image_bytes)
                # t = Timer(10, return None )
                # t.start()
                while True:
                    ready = select.select([self.sock], [], [], 7)
                    if ready[0]:
                        response = self.sock.recv(self.buffer_size)
                        # if response.startswith(b'BOXES'):
                        print('got boxes')
                        # data = response.split(b':')[1]
                        tup = self.processImageLabels(response)
                        # close the connection 
                        self.sock.close()  
                        return tup
                    else:
                        return None
        except Exception as e:
                print("Closed a thread for server", str(e))
        self.sock.close()
        return None

if __name__ == "__main__":
    SingleThreadedClient('', 8000).connectToServer()